from __future__ import absolute_import
from solver4 import Simulation, Expr, Constraint
from circuitsim.ground import Ground

class Circuit(object):
	def __init__(self):
		self._sim = Simulation()
		self._sim.max_inner_timestep = 0.001
		
		self._nodes = set()
		self._components = set()
		
		self._ground_current_constraint = Constraint(Expr.constant(0))
		self._ground_current_constraint.max_iterations = 1
		self._ground_current_constraint.name = "ground current"
		self._sim.add_constraint(self._ground_current_constraint)
		
	def set_debug_file(self, file):
		self._sim.debug_file = file
	
	def _debug(self):
		if self._sim.debug_file:
			print >>self._sim.debug_file, "Circuit %s:" % hex(id(self))
			for component in self._components:
				print >>self._sim.debug_file, "%s:" % repr(component)
				for pin,node in component._pins.iteritems():
					print >>self._sim.debug_file, "\t", "Pin %d:"%pin, hex(id(node))
			
			self._sim.debug(True)
	
	
	def calculate(self):
		# Make sure all components are ready
		for component in self._components:
			for pin in component._pins:
				if component._pins[pin] is None:
					raise Exception, "Component %s has no wire attached to pin %s" % (component,pin)
		
		# Emit debugging information
		self._debug()
		
		knowns = {}
		estimates = {}
		
		# The component and node objects will have assigned 'value' attributes to their variables.
		for var in self._sim.variables:
			if var.derivative:
				knowns[var] = var.value
			else:
				estimates[var] = var.value
		
		variables = self._sim.calculate_completion(knowns, estimates)
		
		for var in self._sim.variables:
			if not var.derivative:
				var.value = variables[var]
	
	def calculate_change(self, t):
		# Make sure all components are ready
		for component in self._components:
			for pin in component._pins:
				if component._pins[pin] is None:
					raise Exception, "Component %s has no wire attached to pin %s" % (component,pin)
		
		# Emit debugging information
		self._debug()
		
		variables = {}
		
		for var in self._sim.variables:
			variables[var] = var.value
		
		variables = self._sim.calculate_change(variables, t)
		
		for var in self._sim.variables:
			var.value = variables[var]
	
	def _update_ground_current_constraint(self):
		# _update_ground_current_constraint rebuilds the constraint on current passing through
		# ground. It is called by Ground objects whenever they are added to the circuit or removed.
		
		grounds = set()
		for component in self._components:
			if isinstance(component, Ground):
				grounds.add(component)
		
		e = Expr.sum(*[Expr.variable(g._current) for g in grounds])
		self._ground_current_constraint.expr = e